#!/local/bin/perl -wT
#
# File: spambounce.pl
#
# Description: first cut at a script intended to bounce a spam message
#              back to someone responsible for the host we received it
#              from.
#
# Author: Sylvain Robitaille
#
# Date: 2003/07/16 (begin)
#       2003/08/06: ensure compatibility with Perl-5.8.0, add query_dns
#       2003/11/25: send complaints using sendmail instead of mailx
#                   add outgoing complaint email headers to
#                   configuration section.
#                   add query_whois, and code to actually send complaint.
#       2003/11/26: Abstract out domain name from hostnames in
#                   mailservers regular expression.  This permits the
#                   headers in the outgoing complaint to use the
#                   configured domain name.
#                   remove some extraneous core from *_message routines.
#                   add verbose/debugging output to query_dns.
#                   add error message if we can't find our locally added
#                   Received: headers.
#                   add a note to contact local abuse contact directly
#                   if spam message was sent from a local system.
#                   debug mode: prints to stdout instead of calling
#                   sendmail.
#       2003/12/01: Sanity checking the SOA contact address.
#       2003/12/03: Tweak the $received regular expression.
#       2003/12/04: Include Received: header in "please contact local
#                   abuse" message;  update history and comments.
#       2003/12/05: Strip out unprintable characters from contact
#                   addresses.
#       2004/05/12: Newsflash is no longer a mail relay.

#
# First some directives for the compiler.
use strict;            # Everybody should use this, IM(NS)HO
use Getopt::Std;       # command line arguments
use File::Basename;    # filename parsing
use Sys::Syslog;       # syslog logging

####################################################################
#
# CONFIGURATION section. Edit to your heart's content!

# The syslog facility to use if running as uid = 0:
my $root_syslog_facility = 'syslog';

# The syslog facility to use if running as any other user. This may or
# may not be the same as $root_syslog_facility.
my $user_syslog_facility = 'user';

# The program to use for mailing errors. Note that this program must
# be able to read its input in order to obtain the address(es) of the
# intended recipient(s).
my $mail_program = 'sendmail -t -oi';

# The subject we'll use when mailing our report.
my $subject = "ABUSE: Unsolicited Bulk email from one of your users";

# Environment variables. Keep these to known safe values.
$ENV{PATH} = '/local/bin:/usr/bin';
$ENV{IFS}  = " \t\n";     # <space><tab><newline>
umask 077;

# Our local domain (regular expression):
my $local_domain = "mydomain\.ca";

# Our mail servers (regular expression):
my $mailservers = "(new)?(smtp1|smtp2|smtp3)\.$local_domain";

# Received headers added by our mail servers (regular expression):
my $received = 'Received:\s+from\s+[\[\]\w\-\_\.]+\s+' .
               '\((?:(?:.*\@)?([\w\-\_\.]+)\s+)?\[([\d\.]+)\]' .
               '(?:\s+\(may\s+be\s+forged\))?\)\s+by\s+' .
                $mailservers .
               '\s+\(([\d\.\/]+(\+\w+)?|MIMEDefang)\)\s+' .
               'with\s+E?SMTP\s+id\s+\S+' .
               '(?:\s+for\s+<[\w\-\_\.]+\@[\w\-\_\.]+>)?;\s+' .
               '\w{3},\s+[\d]{1,2}\s+\w{3}\s+\d{4}\s+\s?[\d:]{7,8}\s+' .
               '-0[45]00(\s+\(E[DS]T\))?';
#               ^^^^^^^^^^^^^^^^^^^^^^^^
# Check the timezone(s) that are reasonable for your area!

# The above regular expression is valid, at least for recent versions
# of Sendmail running on a Unix system.  Adjust as necessary for other
# MTAs and/or other systems.

#
# End of configuration.
####################################################################

#
# Global arrays which are used throughout

#
# "Option" variables.
use vars qw($opt_d $opt_l $opt_v);

# They translate to:
my $debug     = 0;
my $verbose   = 0;
my $listonly  = 0;


# SUBROUTINE: fatal_error( $message )
#
# Description: executes a call to die, with $message as an
#              argument.
#
# NOTE: We want to syslog the error message before dying.
#
# Author: Sylvain Robitaille
#
# Date: 1997/06/19
#       1998/02/04: include logging via syslog.
#
sub fatal_error {
    syslog 'err', "FATAL error: @_\n";
    die "FATAL error: @_\n";
}

# SUBROUTINE: debug_message( $message )
#
# Description: Prints $message to STDOUT, clearly
#              identified as debug information. Prints only 
#              if $debug is a true value.
#
# Author: Sylvain Robitaille
#
# Date: 1997/07/22
#       1998/02/04: include logging via syslog.
#       1998/03/06: but only if in debug mode!
#
sub debug_message {
  if( $debug ) {
    syslog 'debug', "@_\n";
    print "Debug: @_\n"
  }
}

# SUBROUTINE: error_message( $message )
#
# Description: Prints messages to $errorfile if writable, and to
#              STDERR if not in quiet mode.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/02
#       1998/02/04: include logging via syslog.
#       1998/02/06: prefix message with the string 'Error: '
#
sub error_message {

  syslog 'err', "Error: @_\n";
  print STDERR  "Error: @_\n";
}

# SUBROUTINE: info_message( $message )
#
# Description: Prints messages to STDOUT if not in quiet mode.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/04
#
sub info_message {
  syslog 'info', "@_\n";
  print "@_\n" if $verbose;
}

# Subroutine: Usage
#
# Description: Prints a usage message.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/02
#
sub Usage {
  my $progname = basename($0, ());

  print STDERR <<USAGE;
    Usage: $progname [-qld]
USAGE

exit 1;
}

# Subroutine: Parse_Options
#
# Description: Parses the command line options into default global
#              variables.
#
# Uses globals: opt_d
#               opt_q
#
# Author: Sylvain Robitaille
#
# Date:
#
sub Parse_Options {

  if(! getopts('dlv')) {
    Usage;
  }

  $debug       = $opt_d if( $opt_d );
  $listonly    = $opt_l if( $opt_l );
  $verbose     = $opt_v if( $opt_v );
}


use Net::DNS;
# Subroutine: query_dns
#
# Description: Queries DNS for the SOA record of the zone in which
#              $host is found.
#
# Uses globals:
#
# Returns: contact address from SOA record, or undefined.
#
# Author: Sylvain Robitaille
#
# Date: 2003/08/06
#
sub query_dns($) {
    my $host = shift;
    my $res = new Net::DNS::Resolver;
    my $query; 
    my $contact;
    my $origin;
    my $arpa = "";
    
    # If we're working with a numeric IP address:
    if ($host =~ m/^[0-9\.]+$/) {
        # Reverse the IP address order:
        $host =~ s/^([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+)$/$4.$3.$2.$1/;
    
        # Use the .in-addr.arpa. zone:
        $arpa = ".in-addr.arpa.";
    } else {
       # lookup only the domain name?
       $host =~ s%^[^\.]+\.%%;
    }
    
    while(! ($query = $res->query($host . $arpa, "SOA")) ) {
        last if $host !~ m/\./;
        $host =~ s/^[^\.]+\.//;
    }
    
    if ($query) {
       ($origin,$contact) = (($query->answer)[0]->string =~
                                 m/^.+\s+SOA\s+(\S+)\.\s+(\S+)\.\s+\(/);
       $contact =~ s/\.localhost$/\.$origin/;
       #
       # Undefine it if it isn't in the expected format:
       if( $contact !~ m%^\S+$% ) {
          $contact = undef;
       }
    }

    # NOTE: We're returning an undefined value if we didn't find a contact
    #       address.
    return $contact;
}


use Net::Whois::IP qw(whoisip_query);
# Subroutine: query_whois
#
# Description: Queries whois servers for registration of domain
#              responsible for $ip_addr
#
# Uses globals:
#
# Returns: contact address, or undefined.
#
# Author: Sylvain Robitaille
#
# Date: 2003/11/25
#
sub query_whois($) {
    my $ip_addr = shift;
    my $whois_response = whoisip_query($ip_addr, "true");
    my $key;

    # European sites don't seem to have gotten the idea behind having
    # this data be searchable; They seem to think it will only be
    # searched by a human who will read their remarks.  We may have
    # trouble contacting them ...
    #
    # 2004/09/29 Sylvain Robitaille: UPDATE
    # After much arguing back and forth with the folks at RIPE, I'm told
    # that a decision has been reached among their community, and
    # RIPE/NCC (who manages the RIPE data) will add a field for
    # "abuse-mailbox" or something similar.  Until we know what they
    # will call it, let's just start looking for it by that name ...
    foreach $key ("AbuseEmail", "OrgAbuseEmail",
                  "TechEmail",  "OrgTechEmail",
                  "abuse-email", "abuse-mailbox",
                  "notify", "e-mail", "email") {
       if( defined $whois_response->{$key} ) {
          # Remove unprintable characters:
          $whois_response->{$key} =~ s%[\x00-\x1f\x80-\xff]%%sg;
          return $whois_response->{$key}
       }
    }
    # If we got here, we didn't find a contact address, print out what
    # we did find, and return undefined:
    if( $debug || $verbose ) {
        foreach (keys(%{$whois_response}) ) {
           print "$_:";
           foreach ( @{$whois_response->{$_}} ) {
              print "  $_\n";
           }
        }
    }
    return undef;
}


# The main routine.
#
# Author: Sylvain Robitaille
#
# Date: 2003/11/25
#

# Variables used by main.
my $facility;        # Which facility will we log to syslog with?
my $progname;        # Who are we, anyway?
my $oldfh;           # stdout filehandle.

# Now get command line options.
Parse_Options;

# Who are we, anyway?
$progname = basename $0;

# Ok we're started up correctly. We want to open up a connection to
# syslog, with the appropriate facility, and log that we're starting.
#
# We have the following options for syslog logging:
#
# pid:    Logs the process ID with each message.
# ndelay: Opens the connect to syslogd immediately, instead of when the
#         first message is logged.
# cons:   Sends messages to the console if unable to send them to
#         syslogd.
# nowait: Logs messages to the console without waiting for child
#         processes that are forked.
#
# $> is our effective user id.
if( $> == 0 ) {
  $facility = $root_syslog_facility;
} else {
  $facility = $user_syslog_facility;
}
openlog $progname, 'pid', $facility;

# Ok, here we go...
#info_message "$progname started.";

# Read through the message headers:
my $current_header;
my $sending_host;
my $received_header;
my $senders_ip;
my @saved_headers;
while (<>) {
  last if m/^$/;

  chomp;
  if ( m/^\s+/ ) {
      s/^\s+/ /;
      $current_header .= $_;
  } else {
      $current_header  = $_;
  }

  if ( $current_header =~ m/^$received/o ) {
      $sending_host = (defined $1)?$1:$2;
      $senders_ip   = $2;
      $current_header =~ s/(\t+)/\n$1/g;
      $received_header = $current_header;
  }
  push @saved_headers, $_;
}

if( ! defined $sending_host ) {
   error_message "Unable to find expected Received header in the following:";
   error_message "";
   foreach (@saved_headers) {
      error_message $_;
   }
   exit 1;
}

# Ok, we have the original Received: header in $received_header, and
# the sending host in $sending_host.
debug_message $received_header;
debug_message $sending_host;

# print the sending host if we're in "listonly" mode:
if ( $listonly ) {
  print $sending_host . "\n";
} else {
  # If the sending host is "localhost", the message was sent locally:
  if( $sending_host eq "localhost" || $sending_host eq "127.0.0.1" ||
      $sending_host =~ m%\.$local_domain$%i ) {
    error_message "Message was sent from a local system.\n",
                  "Please contact abuse\@$local_domain directly.\n",
                  "\n",
                  $received_header;
    exit 0;
  }
  
  # Next we need to find the contact address for $sending_host, preferably
  # using some sort of whois lookup (or the DNS SOA record?)
  my $dns_contact = query_dns($sending_host);
  if ( defined $dns_contact ) {
     # Replace the first '.' with '@':
     # 2005/01/17 Sylvain Robitaille: only if not already present ...
     $dns_contact =~ s%[\@\.]%\@%;
  } else {
     error_message "can't find SOA record for $sending_host!"
  }
  my $whois_contact = query_whois($senders_ip);
  if ( defined $whois_contact ) {
     $whois_contact = join ',', @{$whois_contact};
  } else {
     error_message "can't find WHOIS contact for $sending_host!"
  }
  
  # If neither $whois_contact nor $dns_contact were found, we have no one
  # to complain to!
  if ( !defined($whois_contact) && !defined($dns_contact) ) {
     fatal_error "can't find a contact address for $sending_host!";
  }
  
  # Finally, we forward the message, as a complaint, to the contact
  # address(es)
  info_message "emailing complaint to",
               defined($whois_contact)?"$whois_contact":"",
               defined($dns_contact)?"$dns_contact":"";
  
  # If we're debugging, we'll simply print to stdout what would be sent
  # via email:
  if( ! $debug ) {
      open OUTPUT, "| $mail_program" or
          fatal_error "can't run $mail_program: $!";
      $oldfh = select OUTPUT;
  } else {
      debug_message "----------------------------------------------------";
  }
  
  if ( defined $whois_contact ) {
     print "To: " . $whois_contact . "\n";
     if ( defined $dns_contact ) {
        print "Cc: " . $dns_contact . "\n";
     }
  } else {
     if ( defined $dns_contact ) {
        print "To: " . $dns_contact . "\n";
     }
  }
  print "From: Spam Complaint <postmaster\@$local_domain>\n",
        "Subject: $subject\n",
        "\n",
        "A message containing the following headers was sent to one or\n",
        "more of our users via one of your systems. It is unsolicited\n",
        "bulk email and is not appreciated nor desired.\n",
        "\n",
        "Please make sure your customers do not use your systems\n",
        "to abuse services in this way.\n",
        "\n";
  
  foreach (@saved_headers) {
      print $_ . "\n";
  }
  if( ! $debug ) {
      select $oldfh;
      close OUTPUT;
  }
}

# Log that we're done before closing the connectio to syslog
#info_message "$progname done.";
closelog;
